% [xk,niter,residuals,outputData,opts] =Core_Nesterov_UP(A,At,b,Lambda,La,mu,opts)
%
% Solves a L1 or TV minimization + quadratic termproblem
% Nesterov algorithm, without continuation:
%
%     min_x lambda || U x ||_1 + 1/2 ||b - Ax||_2^2 
% 
% If continuation is desired, see the function NESTA.m
%
% The primal prox-function is also adapted by accounting for a first guess
% xplug that also tends towards x_muf 
%
% The observation matrix A need not be a projector
%
% Inputs:   A and At - measurement matrix and adjoint (either a matrix, in which
%               case At is unused, or function handles)
%           b   - Observed data, a m x 1 array
%           Lambda - Lagrange multiplier
%               Common heuristic: Lambda = sigma*sqrt(2*log(n)),
%               where sigma=std(noise).
%           La - Lipschitz constant of the quadratic term; La =  ||A||^2
%           muf - The desired value of mu at the last continuation step.
%               A smaller mu leads to higher accuracy.
%           opts -
%               This is a structure that contains additional options,
%               some of which are optional.
%               The fieldnames are case insensitive.  Below
%               are the possible fieldnames:
%               
%               opts.xplug - the first guess for the primal prox-function, and
%                 also the initial point for xk.  By default, xplug = At(b)
%               opts.U and opts.Ut - Analysis/Synthesis operators
%                 (either matrices of function handles).
%               opts.normU - if opts.U is provided, this should be norm(U)
%               opts.maxiter - max number of iterations in an inner loop.
%                 default is 10,000
%               opts.TolVar - tolerance for the stopping criteria
%               opts.stopTest - which stopping criteria to apply
%                   opts.stopTest == 1 : stop when the relative
%                       change in the objective function is less than
%                       TolVar
%                   opts.stopTest == 2 : stop with the l_infinity norm
%                       of difference in the xk variable is less
%                       than TolVar
%               opts.TypeMin - if this is 'L1' (default), then
%                   minimizes a smoothed version of the l_1 norm.
%                   If this is 'tv', then minimizes a smoothed
%                   version of the total-variation norm.
%                   The string is case insensitive.
%               opts.Verbose - if this is 0 or false, then very
%                   little output is displayed.  If this is 1 or true,
%                   then output every iteration is displayed.
%                   If this is a number p greater than 1, then
%                   output is displayed every pth iteration.
%               opts.fid - if this is 1 (default), the display is
%                   the usual Matlab screen.  If this is the file-id
%                   of a file opened with fopen, then the display
%                   will be redirected to this file.
%               opts.errFcn - if this is a function handle,
%                   then the program will evaluate opts.errFcn(xk)
%                   at every iteration and display the result.
%                   ex.  opts.errFcn = @(x) norm( x - x_true )
%               opts.outFcn - if this is a function handle, 
%                   then then program will evaluate opts.outFcn(xk)
%                   at every iteration and save the results in outputData.
%                   If the result is a vector (as opposed to a scalar),
%                   it should be a row vector and not a column vector.
%                   ex. opts.outFcn = @(x) [norm( x - xtrue, 'inf' ),...
%                                           norm( x - xtrue) / norm(xtrue)]
%  Outputs:
%           xk  - estimate of the solution x
%           niter - number of iterations
%           residuals - first column is the residual at every step,
%               second column is the value of f_mu at every step
%           outputData - a matrix, where each row r is the output
%               from opts.outFcn, if supplied.
%           opts - the options that were used
%
% Written by: Jerome Bobin, Caltech
% Email: bobin@acm.caltech.edu
% Created: May 2009
% Modified: Nov 2009 Stephen Becker
%
% NESTA Version 1.1
%   See also NESTA_UP


function [xk,niter,residuals,outputData,opts,xk_log] = Core_Nesterov_UP(A,At,b,Lambda,La,mu,opts)

%---- Set defaults
%opts = [];
if nargin < 7, opts = []; end
if isempty(opts) && isnumeric(opts), opts = struct; end
fid = setOpts('fid',1);
function printf(varargin), fprintf(fid,varargin{:}); end
maxiter = setOpts('maxiter',10000,0);
TolVar = setOpts('TolVar',1e-5);
TypeMin = setOpts('TypeMin','L1');
Verbose = setOpts('Verbose',true);
errFcn = setOpts('errFcn',[]);
outFcn = setOpts('outFcn',[]);
stopTest = setOpts('stopTest',1,1,2);
miniter = setOpts('miniter',7,1,1000); %%%
U = setOpts('U', @(x) x );
if ~isa(U,'function_handle')
    Ut = setOpts('Ut',[]);
else
    Ut = setOpts('Ut', @(x) x );
end
normU = setOpts('normU',1);
xplug = setOpts('xplug',[]);
if isempty(xplug)
    if isa(A,'function_handle')
        xplug=At(b);
    else
        xplug = A'*b;
    end
end

%--- Trace log options ---
%  --- Javier Royuela <jroyval@lpi.tel.uva.es>

LogFrames     = setOpts('logFrames',0);
LogIterPeriod = setOpts('logIterPeriod',0);
DataShape     = setOpts('dataShape',size(xplug));
klog = 0;
if LogIterPeriod == 0
    LOGEVOL = 0;
else
    LOGEVOL = 1;
    FrameSize = DataShape(1)*DataShape(2);
    
    % Selection index vector to extract only desired frames
    selvec = bsxfun(@plus,(1:FrameSize)',FrameSize*(LogFrames(:)'-1));
    selvec = selvec(:);
    
    xk_log = zeros(maxiter/LogIterPeriod + 1,length(selvec));
end

%---- Initialization
N = length(xplug);
wk = zeros(N,1); 
xk = xplug;

%---- Init Variables
Ak= 0;
Lmu = normU/mu;
yk = xk;
zk = xk;
f_buff = Inf*ones(1, miniter); %%%
xk_buf = zeros(size(xk,1), miniter);      %%%
xknorm_buf = zeros(1, miniter);  %%%
k_last = 1; %%%

OK = 0;
n = floor(sqrt(N));

%---- TV Minimization
if strcmp(TypeMin,'tv')
    Lmu = 8*Lmu;
    global T_frames ROW COL
    if isempty(T_frames)
        T_frames = 1;
        N = length(x)/T_frames; n = floor(sqrt(N));
        ROW = n; COL = n;
    end
    N = ROW*COL;
    Dv = spdiags(repmat([reshape([-ones(ROW-1,COL); zeros(1,COL)],N,1) ...
        reshape([zeros(1,COL); ones(ROW-1,COL)],N,1)], T_frames,1), [0 1], T_frames*N, T_frames*N);
    Dh = spdiags(repmat([reshape([-ones(ROW,COL-1) zeros(ROW,1)],N,1) ...
        reshape([zeros(ROW,1) ones(ROW,COL-1)],N,1)],T_frames,1), [0 ROW], T_frames*N, T_frames*N);
    D = sparse([Dh;Dv]);
    
    %     Dv = spdiags([reshape([-ones(n-1,n); zeros(1,n)],N,1) ...
    %         reshape([zeros(1,n); ones(n-1,n)],N,1)], [0 1], N, N);
    %     Dh = spdiags([reshape([-ones(n,n-1) zeros(n,1)],N,1) ...
    %         reshape([zeros(n,1) ones(n,n-1)],N,1)], [0 n], N, N);
    %     D = sparse([Dh;Dv]);
end

Lmu = Lambda*Lmu + La;

Lmu1 = 1/Lmu;
SLmu = sqrt(Lmu);
SLmu1 = 1/sqrt(Lmu);

%---- setup data storage variables
[DISPLAY_ERROR, RECORD_DATA] = deal(false);
outputData = deal([]);
residuals = zeros(maxiter,2);
if ~isempty(errFcn), DISPLAY_ERROR = true; end
if ~isempty(outFcn) && nargout >= 4
    RECORD_DATA = true;
    if ishandle(outFcn)
        outputData = zeros(maxiter, size(outFcn(xplug),2) );
    else
        outputData = zeros(maxiter, 1);
    end
end

for k = 0:maxiter-1,
    
   %---- Dual problem
   
   if strcmp(TypeMin,'L1')
       if strcmp(outFcn, 'L1')
           [df,fx,val,uk, uk_l1] = Perform_L1_Constraint(xk,mu,U,Ut);
       else
           [df,fx,val,uk] = Perform_L1_Constraint(xk,mu,U,Ut);
       end
   end
   
   if strcmp(TypeMin,'tv')
       [df,fx] = Perform_TV_Constraint(xk,mu,Dv,Dh,D,U,Ut);
   end
   
   if isa(A,'function_handle'),
        res = A(xk) - b;
        Ares = At(res);
   else
       res = A*xk - b;
       Ares = A'*(res);
   end
   
   df = Lambda*df + Ares;
   
   %---- Modification to log l2term and l1term evolution.
   % Javier Royuela <jroyval@lpi.tel.uva.es>
   l2term = sum(abs(res(:).^2));
   l1term = fx;
   fx = 0.5*l2term + Lambda*l1term;
   %---- end Mod.
   
   %---- Updating yk 
    
   yk = xk - Lmu1*df;
    
   residuals(k+1,1) = gather(l2term);    % the l2 residual
   residuals(k+1,2) = gather(l1term);    % the value of l1term
   
    %--- if user has supplied a function, apply it to the iterate
    if RECORD_DATA
        if ishandle(outFcn)
            outputData(k+1,:) = outFcn(xk);
        elseif strcmp(outFcn,'L1');
            outputData(k+1,:) = uk_l1;
        end
    end
    
    %---# Logging modification J.R.
    if LOGEVOL
        if mod(k,LogIterPeriod)==0
            klog = klog + 1;
            xk_log(klog,:) = xk(selvec);
        end
    end
    %---# Modification end
    
    %--- Stopping criteria
    
    % miniter = 7; %%% -> Now in opts.miniter J.R.
    % xknorm == mean value 
    xknorm = norm(xk)/sqrt(numel(xk)); %%% rms pixel intensity
    switch stopTest
        case 1
            % look at the relative change in function value
            qp = (mean(f_buff) - fx)/mean(f_buff);
            
            if fx > f_buff(k_last)
                warning('Function is increassing: %f -> %f', f_buff(k_last), fx);
            end
            
            if k > miniter && qp <= TolVar || fx > f_buff(k_last);
                fprintf('Done. qp = %e\n', qp);
                break;
            end
            
            f_buff(k_last) = fx;

        case 2
            % look at the l_inf change from solution miniter iterations
            % before
            
            if fx > f_buff(k_last)
                warning('Function is increassing: %f -> %f', f_buff(k_last), fx);
            end
            
            if k>=1
                qp = norm( xk - xk_buf(:,k_last), 'inf' )/xknorm_buf(k_last);
            else
                qp = Inf;
            end
            
            if k >= miniter && qp <= TolVar
                fprintf('Done. qp = %e\n', qp);
                break;
            end
            
            xk_buf(:,k_last) = xk;
            xknorm_buf(k_last) = xknorm;
            f_buff(k_last) = fx;
    end
    
    k_last = mod(k, miniter) + 1;

    %--- Updating zk
  
    apk = 0.5*(k+1);
    Ak = Ak + apk; 
    tauk = 2/(k+3); 
    
    wk =  apk*df + wk;
    zk = xplug - Lmu1*wk;

    %--- Updating xk
    
    xkp = tauk*zk + (1-tauk)*yk;
    xk=xkp; 
    
    %--- display progress if desired
    if ~mod(k+1,Verbose )
        printf(': %3d | f: %.3e | qp: %.3e | l2: %.3e | l1: %.3e | xk: %.3e',...
            k+1,fx,qp,residuals(k+1,1),residuals(k+1,2),xknorm); 
        %--- if user has supplied a function to calculate the error,
        % apply it to the current iterate and dislay the output:
        if DISPLAY_ERROR, printf(' ~ Error: %.2e',errFcn(xk)); end
        printf('\n');
        
        %         %%%%%%%%%%%%%%%%%%%%%%%MOTION SPECIFIC%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %         global T_frames ROW COL
        %         Ir_cube = [];
        %         for frame = 2:T_frames-1;
        %             Ir_cube(:,:,frame) = reshape(xk((frame-1)*ROW*COL+1:frame*ROW*COL),ROW,COL);
        %             figure(333); imagesc(abs(Ir_cube(:,:,frame)));
        %             axis image; colormap gray; pause(1/60);
        %         end
        
%        diary OFF; diary ON;
    end

end

niter = k+1; 

residuals = residuals(1:niter,:);
if RECORD_DATA
    outputData = outputData(1:niter,:);
end

if LOGEVOL
    xk_log = xk_log(1:klog,:);
else
    xk_log = [];
end
%end

%---- internal routine for setting defaults
    function var = setOpts(field,default,mn,mx)
    var = default;
    % has the option already been set?
    if ~isfield(opts,field) 
        % see if there is a capitalization problem:
        names = fieldnames(opts);
        for i = 1:length(names)
            if strcmpi(names{i},field)
                opts.(field) = opts.(names{i});
                opts=rmfield(opts,names{i});
                break;
            end
        end
    end

    if isfield(opts,field) && ~isempty(opts.(field))
        var = opts.(field);  % override the default
    end

    % perform error checking, if desired
    if nargin >= 3 && ~isempty(mn)
        if var < mn
            printf('Variable %s is %f, should be at least %f\n',...
                field,var,mn); error('variable out-of-bounds');
        end
    end
    if nargin >= 4 && ~isempty(mx)
        if var > mx
            printf('Variable %s is %f, should be at least %f\n',...
                field,var,mn); error('variable out-of-bounds');
        end
    end
    opts.(field) = var;

end

end %% end of main Core_Nesterov routine

%%%%%%%%%%%% PERFORM THE L1 CONSTRAINT %%%%%%%%%%%%%%%%%%

function [df,fx,val,uk,l1_k] = Perform_L1_Constraint(xk,mu,U,Ut)

    if isa(U,'function_handle')
        uk = U(xk);
    else
        uk = U*xk;
    end

    fx = uk;
    
    % Get true l1 norm for logging purposes
    if nargout==5
        l1_k = sum(abs(uk));
    end
        
    uk = uk./max(mu,abs(uk));
    val = real(uk'*fx);
    fx = real(uk'*fx - mu/2*norm(uk)^2);
    
    if isa(Ut,'function_handle')
        df = Ut(uk);
    else
        df = U'*uk;
    end

end

%%%%%%%%%%%% PERFORM THE TV CONSTRAINT %%%%%%%%%%%%%%%%%%

function [df,fx] = Perform_TV_Constraint(xk,mu,Dv,Dh,D,U,Ut)

    if isa(U,'function_handle')
        x = U(xk);
    else
        x = U*xk;
    end
    df = zeros(size(x));
    
    Dhx = Dh*x;
    Dvx = Dv*x;
    
    tvx = sum(sqrt(abs(Dhx).^2+abs(Dvx).^2));
    w = max(mu,sqrt(abs(Dhx).^2 + abs(Dvx).^2));
    uh = Dhx ./ w;
    uv = Dvx ./ w;
    u = [uh;uv];
    fx = u'*D*x - mu/2 * norm(u)^2;
    if isa(Ut,'function_handle')
        df = Ut(D'*u);
    else
        df = U'*(D'*u);
    end

end
