function [ts, T] = bspline_init(x,box,Dp,E,N)
% BSPLINE_INIT Initialize 2D+t B-spline interpolation structure.
%   TS = BSPLINE_INIT(X,BOX,DP,E,N) returns TS, a structure for
%   permforming B-spline interpolation with fields:
%      ts.BB: B-spline kernel
%      ts.v1: Control points horizontal dimension coordinates
%      ts.v2: Control points vertical dimensino coordinates
%      ts.Dp: Control points spacing on horizontal and vertical directions
%      ts.E:  B-spline order
%   [TS, T] = BSPLINE_INIT(X,BOX,DP,E,N) returns TS and T, an initial
%   zero-valued control points matrix.
%   INPUT arguments:
%      x  : image coordinates grid (meshgrid)
%      box: limits of the box containing the ROI
%      Dp : control points spacing Dp=[D_horiz D_vert];
%      E  : B-spline order
%      N  : number of temporal frames
%
%   Copyright (C) 2015 Javier Royuela del Val  <jroyval@lpi.tel.uva.es>
%                      Santiago Sanz Estébanez <ssanest@lpi.tel.uva.es>
%   Laboratorio de Procesado de Imagen, Universidad de Valladolid
%   www.lpi.tel.uva.es
%
%   This program is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
%
%   This program is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with this program.  If not, see <http://www.gnu.org/licenses/>.

    % Size of the image
    tamxy = [size(x,2) size(x,1)];

    % Number of spatial dimensions
    L = size(x,3); 

    % Central pixel coordinates (rounded)
    c = floor( (1 + tamxy(1:2)) / 2);

    % B-splines convolution kernel.
    %  - Finite support given by its order and control points spacing.
    %  - Zero values at the edges are omitted.
    xh = ( -(E+1)/2 + 1/Dp(1) : 1/Dp(1) : (E+1)/2 - 1/Dp(1) );
    yh = ( -(E+1)/2 + 1/Dp(2) : 1/Dp(2) : (E+1)/2 - 1/Dp(2) )';
 
%     [xxh, yyh] = meshgrid(xh, yh);
    
    BB = bsplineN(yh, E) * bsplineN(xh, E);
    
    dBBdx2 = dbsplineNp(yh,E,1) * bsplineN(xh, E);
    dBBdx1 = bsplineN(yh, E)    * dbsplineNp(xh,E,1);
    
    d2BBdx2    = dbsplineNp(yh,E,2) * bsplineN(xh, E);
    d2BBdx1    = bsplineN(yh, E)    * dbsplineNp(xh,E,2);
    d2BBdx1dx2 = dbsplineNp(yh,E,1) * dbsplineNp(xh,E,1);

    % Gradient computation:
    %  - Each control point affects its neighbourhood. Its radus is given
    %  by (E+1)*Dp(:)/2.
    rp=(E+1)*Dp(:)/2;

    %  Control points positions:
    %  - Inside box containing ROI + margin (rp).
    %  - One control point in center of the image (c)
    c_ul = box(:,1)-rp;  % up-left border
    c_dr = box(:,2)+rp;  % down-right border
    
    % To the left and upside, departing from center:
    v1 = fliplr( c(1) : -Dp(1) : c_ul(1) ); 
    v2 = fliplr( c(2) : -Dp(2) : c_ul(2) );
    % To the right and downside, departing from center + Dp:
    v1 = [ v1, c(1)+Dp(1) : Dp(1) : c_dr(1) ];
    v2 = [ v2, c(2)+Dp(2) : Dp(2) : c_dr(2) ];
 
% % % No necesario meshgrid si puntos en reticula
% % % [v(:,:,1) v(:,:,2)] = meshgrid(v1,v2);
% % % Puedo asignar/recuperar los desplazamientos de los puntos de control con
% % %    T(v1,v2,nt,l) = ...;
% % % La matriz de puntos de control:
% % %   - tiene el mismo tamaño que la secuencia original, size(T)==[size(I), L]
% % %   - valdra cero en todo punto menos en punto de control
% % %   - no se puede hacer sparse por algoritmo de convolucion
% % %        ** en mathworks hay implementacion de conv2 para matrices sparse
% % % T = zeros([size(I) L]);
    
    % Control points displacement field init matrix:
    dataType = class(x);

    if nargout > 1
        T = zeros([length(v2) length(v1) N L], dataType);
    end

    ts.BB = BB;
    ts.v1 = v1;
    ts.v2 = v2;
    ts.Dp = Dp;
    ts.E  = E;
    
    ts.dBBdx1 = dBBdx1;
    ts.dBBdx2 = dBBdx2;
    
    ts.d2BBdx1 = d2BBdx1;
    ts.d2BBdx2 = d2BBdx2;
    ts.d2BBdx1dx2 = d2BBdx1dx2;
end

