function [E, dE, Ex] = elastpenalty(x,xn,T,ts,Wn,lambda,p) %#ok<INUSL>
%
%   Copyright (C) 2015 Javier Royuela del Val  <jroyval@lpi.tel.uva.es>
%                      Santiago Sanz Estébanez <ssanest@lpi.tel.uva.es>
%   Laboratorio de Procesado de Imagen, Universidad de Valladolid
%   www.lpi.tel.uva.es
%
%   This program is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
%
%   This program is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with this program.  If not, see <http://www.gnu.org/licenses/>.


if ~exist('p','var') || isempty(p) || p < 1 || p > 3
	
end

if all(lambda==0)
    E  = 0;
    dE = 0;
    Ex = 0;
    return;
end

s=size(xn);
dataType = class(x);
  
% Spatio-temporal regularization
% 1st spatial
if lambda(1) > 0
    [dxndx1, dxndx2] = bspline_spatial_gradient(x, ts, T, 1);
else
    dxndx1 = zeros(s,dataType);
    dxndx2 = zeros(s,dataType);
end
% 2nd spatial
if lambda(2) > 0
    [d2xndx1, d2xndx2, d2xndx1dx2] = bspline_spatial_gradient(x, ts, T, 2);
else
    d2xndx1    = zeros(s,dataType);
    d2xndx2    = zeros(s,dataType);
    d2xndx1dx2 = zeros(s,dataType);
end
%1st temporal
if lambda(3)>0
    dxndt = bspline_temporal_gradient(xn, 1);
else
    dxndt = zeros(s,dataType);
end
%2nd temporal
if lambda(4)>0
    d2xndt= bspline_temporal_gradient(xn, 2);
else
    d2xndt = zeros(s,dataType);
end

if p<=2
    Ex = lambda(1)*sum( ( 0 + dxndx1 ) .^ 2 + ( 0 + dxndx2 ).^2, 4 ) + ...
         lambda(2)*sum( d2xndx1 .^ 2 + d2xndx2 .^ 2 + 2 * d2xndx1dx2 .^ 2, 4) +...
         lambda(3)*sum( dxndt .^ 2, 4 ) + ...
         lambda(4)*sum( d2xndt .^ 2, 4 );
    E  = sum(Ex(:));
else
    Ex = zeros(s(1:3),dataType);
    E  = 0;
end

E = gather(E);
  
% Regularization gradient with respect to control points:
dE = zeros(size(T),dataType);
if p>=2
    % 1st temporal
    if lambda(3) > 0
      % 2nd order temporal diff needed
      dxndt = diff(xn(:,:,[end 1:end 1],:),2,3);
    end
    % 2nd temporal  
    if lambda(4) > 0
      % 4th order temporal diff needed
      d2xndt = diff(xn(:,:,[end-1 end 1:end 1 2],:),4,3);
    end
    
    v1 = ts.v1 - ts.v1(1) + 1;
    v2 = ts.v2 - ts.v2(1) + 1;
    n1 = ts.v1(1):ts.v1(end);
    n2 = ts.v2(1):ts.v2(end);

    N = s(3);
    L = s(4);
    
    % First order spatial
    if lambda(1) > 0
        aux = convn(dxndx1(n2,n1,:,:),2*ts.Dp(1)*ts.dBBdx1,'same');
        dE = dE + lambda(1)*aux(v2,v1,:,:);
        aux = convn(dxndx2(n2,n1,:,:),2*ts.Dp(2)*ts.dBBdx2,'same');
        dE = dE + lambda(1)*aux(v2,v1,:,:);
    end
    % Second order spatial
    if lambda(2) > 0
        aux = convn(d2xndx1(n2,n1,:,:),2*ts.Dp(1).^2*ts.d2BBdx1,'same');
        dE = dE + lambda(2)*aux(v2,v1,:,:);
        aux = convn(d2xndx2(n2,n1,:,:),2*ts.Dp(2).^2*ts.d2BBdx2,'same');
        dE = dE + lambda(2)*aux(v2,v1,:,:);
        aux = convn(d2xndx1dx2(n2,n1,:,:),2*ts.Dp(1)*ts.Dp(2)*ts.d2BBdx1dx2,'same');
        dE = dE + lambda(2)*aux(v2,v1,:,:);
    end
    % First order temporal
    if lambda(3) > 0
        aux = convn(dxndt(n2,n1,:,:),-2*ts.BB,'same');
        dE = dE + lambda(3)*aux(v2,v1,:,:);
    end          
    % Second order temporal
    if lambda(4) > 0
        aux = convn(d2xndt(n2,n1,:,:),2*ts.BB,'same');
        dE = dE + lambda(4)*aux(v2,v1,:,:);
    end

end
  
E = gather(E);
dE = gather(dE);
Ex = gather(Ex);

end