function [Ir_mc, trace] = gwcs_function(data, csm, opts)
%gwcs_function GW-CS reconstruction of MRI data
%	out = gwcs_function(data, csm, opts)
%
%	Copyright (C) 2015 Javier Royuela del Val <jroyval@lpi.tel.uva.es>
%	Laboratorio de Procesado de Imagen, Universidad de Valladolid
%	www.lpi.tel.uva.es
%
%   When indicated, some parts are based on code from Salman Asif:
%       http://users.ece.gatech.edu/sasif/dynamicMRI/index.html
%
%   This program is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
%
%   This program is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with this program.  If not, see <http://www.gnu.org/licenses/>.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Unpack some options from struct
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

rf             = opts.rf;

INIT_SPARSE    = opts.init_sparse;
SPAT           = opts.spat;
TEMP           = opts.temp;

lambda_i       = opts.li;
lambda_t       = opts.lt;
lambda_s       = opts.ls;

MOTION_ITERS   = opts.motion_iters + 1;
MOT_METRIC     = opts.mot.metric;
MOT_WEIGHTS    = opts.mot.weights;

CLINES         = opts.subsamp.clines;
RSEED          = opts.subsamp.rseed;
SIGMA          = opts.subsamp.sigma;

SAVE_RESULTS   = opts.out.save;
SAVE_TRACES    = opts.out.trace;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Normalize csm to get sum(|c_i|^2) = 1
%  (CG-SENSE preconditioning)
csm.csm = lpi_correct_csm_rms(csm.csm, csm.rms);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Get ground truth image form fully sampled data:
I_ktoi = ktoi(data, csm);
% Correct for non-homogeneous coverage (regard CG-SENSE preconditioning)
I_full = lpi_correct_csm_rms(I_ktoi, csm.rms);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Sensing
%    Retrospective undersampling of the k-space data.
%    Phase enc. direction along first dimension (ROW).
%    Freq. enc. direction along second dimension (COL).
%    Reconstruction matrix size == fully sampled k-space data dimensions.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

[ROW, COL, N, COILS] = size(data);
mat_size = [ROW, COL, N, COILS];

% Set MATLAB random generator seed
try
    rng(RSEED);
catch
    rand('seed',RSEED); %#ok<RAND>
    fprintf('rng(rseed) error. Use rand(''state'')\n');
end

%SAMP_MASK   = gauss_subsampling(ROW, N, rf, CLINES, SIGMA);
SAMP_MASK   = Hybrid_DownsamplingMASK_mod(ROW, COL, N, rf, CLINES);

SAMP_MASK   = logical(ifftshift(SAMP_MASK,1));

N_kt      = sum(SAMP_MASK, 1); % Number of lines acquired per cardiac phase
N_kt_acum = [0 cumsum(N_kt)];  % Accumulated number of k-space lines

% Reallocate undersampled data
y = zeros([N_kt_acum(end) COL COILS]);
for frame = 1:N
    y(N_kt_acum(frame)+1:N_kt_acum(frame+1),:,:) = data(SAMP_MASK(:,frame),:,frame,:);
end

% Aply ifft in readout direction. This way, computational cost gets
% reduced: only 1D fft in phase encoding direction has to be computed by
% A operator in || y - A m ||_2^2 term
y = ifft(y,[],2) * sqrt(COL);

% Normalize data (and fully sampled solution to compute SER)
y_rms  = sqrt(mean(abs(y(:).^2)));
y      = y(:)/y_rms;
I_full = I_full/y_rms;
I_ktoi = I_ktoi/y_rms;

clear data;

% Measure operator
A_meas  = @(z)  A_L2(z, mat_size, SAMP_MASK, csm.csm);
At_meas = @(z) At_L2(z, mat_size, SAMP_MASK, csm.csm);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Set wavelet parameters. From Asif's Code
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if any(strcmpi({SPAT, INIT_SPARSE}, 'spCWT'))
    J = floor(log2(min(ROW,COL)))-4; % Number of scales in wavelet transform
    % Complex wavelets
    % (Fhh +/- Fgg, Fhg +/- Fgh)
    SYM = 3;
    [Faf, Fsf, af, sf] = BiOrthDualFilt_mod;
    C2D = 0;    % 1 -- complex coefficients
                % 0 -- real and imaginary separate
    psiT = @(z) CWT2D_op(z, Faf, af, Fsf, sf, J, SYM, C2D, ROW, COL, N);
    psi  = @(z) adj_CWT2D_op(z, Faf, af, Fsf, sf, J, SYM, C2D, ROW, COL, N);
else
    psiT = @(z)[];
    psi  = @(z)[];
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% GROUPWISE REGISTRATION SETUP
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Original coordinates grid
clear x;
[x(:,:,1), x(:,:,2)] = meshgrid(1:COL,1:ROW);

ROI = opts.mot.roi;  % ROI where metric is computed
USE_GPU_REG      = false;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% START RECONSTRUCTION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Prepare some outputs
SER_stack        = cell(1, 4*MOTION_ITERS + 1);
SER_stack{1, 1}  = rf;
SSIM_stack       = cell(1, 4*MOTION_ITERS + 1);
SSIM_stack{1, 1} = rf;
resid_stack      = cell(1, MOTION_ITERS);
outData_stack    = cell(1, MOTION_ITERS);
lambdavali_stack = zeros(1, MOTION_ITERS);
    
regout.T_ite = [];

cube = @(x)reshape(x,[ROW COL N]);

tic_total = tic();
for m_iter = 1:MOTION_ITERS

    % Define L1 operator:
    if m_iter > 1 
        % MC reconstruction
       
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % GROUPWISE REGISTRATION 
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        % Register last reconstruction:
        fprintf('Registration\n');
        regout = groupwise_reg(Ir_mc, regout.T_ite, MOT_METRIC, MOT_WEIGHTS, USE_GPU_REG, ROI);
        fprintf(1,' registration finished in %d seconds.\n', regout.exectime);
        exectimes.registration(m_iter) = regout.exectime;
        
        if strcmpi(TEMP,'tTVJac')
            % Invert motion fields
            fprintf('Invert motion fields\n'); tic;
            xninv = inv_motion_field(regout.xn);
            exectimes.invertfields = toc;
            % Define operators:
            sm = [ROW, COL, N];
        
            [xn_back, xn_forw] = getBackAndForwardMotionFields(x, regout.xn, xninv);
            A_motion =  @(z) A_L1_MC(z, x, xn_back, 'cubic', lambda_s, sm, SPAT, TEMP, psiT, xn_forw);
            At_motion = @(z)At_L1_MC(z, x, xn_back, 'cubic', lambda_s, sm, SPAT, TEMP, psi , xn_forw);

        else
            A_motion =  @(z) A_L1_MC(z, x, regout.xn, 'cubic', lambda_s, sm, SPAT, TEMP, psiT, []);
            At_motion = @(z)At_L1_MC(z, x, regout.xn, 'cubic', lambda_s, sm, SPAT, TEMP, psi , []);
        end
                
    else
        % Initial reconstruction without motion correction
        
        switch INIT_SPARSE
            case 'spCWT' % Spatial DT-CWT
                A_motion = @(z) psiT(z);
                At_motion= @(z) psi(z);
            case 'tTV'  % Temporal TV
                A_motion = @(z) tTV(z,[ROW, COL, N]);
                At_motion= @(z) tTVt(z,[ROW, COL, N]);
            case 'spTV' % Spatial TV
                A_motion = @(z) spTV(z, [ROW COL N]);
                At_motion= @(z) spTVt(z, [ROW COL N]);
            otherwise % Default option: spatial DT-CWT
                A_motion = @(z) psiT(z);
                At_motion= @(z) psi(z);
        end


    end
        
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % NESTA solver options
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    optsf = struct();
    optsf.Verbose = 1;

    optsf.MaxIntIter = 5; 
    optsf.maxiter    = 100;
    optsf.TolVar     = 3e-4;
    optsf.stoptest   = 1;
    optsf.miniter    = 7;
    mu_f = 3e-4;

    % Initialization
    if m_iter == 1
        optsf.xplug = At_meas(y);
        Ir_zero = lpi_correct_csm_rms(cube(optsf.xplug), csm.rms);
    else
        optsf.xplug = xk_NESTA;
    end
    
    % Logging options
    optsf.logFrames = [1 8 16];
    optsf.logIterPeriod = 5;
    % optsf.logIterPeriod = 0;
    optsf.dataShape =[ROW COL N];
    
    % Measure operator norm <= 1
    La = 1;
 
    % Operators
    optsf.U  = A_motion;
    optsf.Ut = At_motion;
    optsf.typemin = 'L1';

    optsf.errFcn = @(x) norm( abs(x(:)) - abs(I_ktoi(:))) / norm(I_ktoi(:));
    optsf.outFcn = 'L1';
    
    if m_iter == 1
        lambda = lambda_i;
    else
        lambda = lambda_t;
    end
    
    % Final value of lambda and mu are scaled according to dynamic range of the
    % solution and data norm:
    lambdavali = lambda*max(abs(optsf.xplug));
    mu_f_0 = mu_f * max(abs(optsf.xplug));
    
    ta = tic();
    fprintf('Solving unconstrained NESTA using muf=%3.4g, lambda=%3.4g, La=%1.2g, maxiter=%d...\n', ...
        mu_f_0,lambdavali,La,optsf.maxiter);
    [xk_NESTA, ~, resid, outData, ~, xklog] = ...
        NESTA_UP(A_meas,At_meas,y,lambdavali, La, mu_f_0, optsf);
    exectimes.reconstruction(m_iter) = toc(ta);
    
    % Correct for non-homogenities in coil profiles (CG-SENSE precond)
    Ir_mc = lpi_correct_csm_rms(cube(xk_NESTA), csm.rms);
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Compute SER and SSIM
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    normvec    = @(x) norm(col(abs(x)));
    normvecroi = @(x) normvec(x(opts.roi.ver,opts.roi.hor,:));
        
    SER     = normvec(I_full)^2/normvec(abs(Ir_mc)-abs(I_full))^2;
    SER_ROI = normvecroi(I_full)^2/normvecroi(abs(Ir_mc)-abs(I_full))^2;

    SER_frame     = zeros(N,1);
    SER_ROI_frame = zeros(N,1);
    for i=1:N
        SER_frame(i) = normvec(I_full(:,:,i))^2/normvec(abs(I_full(:,:,i))-abs(Ir_mc(:,:,i)))^2;
        SER_ROI_frame(i) = normvecroi(I_full(:,:,i))^2/normvecroi(abs(I_full(:,:,i))-abs(Ir_mc(:,:,i)))^2;
    end

    SER_stack{1,4*(m_iter-1)+2} = SER;
    SER_stack{1,4*(m_iter-1)+3} = SER_ROI;
    SER_stack{1,4*(m_iter-1)+4} = SER_frame;
    SER_stack{1,4*(m_iter-1)+5} = SER_ROI_frame;
    
    % Compute SSIM values
    L = squeeze(max(max(abs(I_full)))); % Max of each frame
    SSIM_frame     = zeros(N,1);
    SSIM_ROI_frame     = zeros(N,1);
    if opts.ssim.maps
        error('mapas no implementados. ejecuta sin ellos');
    else
        for i=1:N
            [SSIM_frame(i),     ~ ] = ssim_index(abs(Ir_mc(:,:,i)),abs(I_full(:,:,i)),opts.ssim.K,opts.ssim.window,L(i));
            [SSIM_ROI_frame(i), ~ ] = ssim_index(abs(Ir_mc(opts.roi.ver,opts.roi.hor,i)),abs(I_full(opts.roi.ver,opts.roi.hor,i)),opts.ssim.K,opts.ssim.window,L(i));
        end
    end

    SSIM_stack{1,4*(m_iter-1)+2} = mean(SSIM_frame);
    SSIM_stack{1,4*(m_iter-1)+3} = mean(SSIM_ROI_frame);
    SSIM_stack{1,4*(m_iter-1)+4} = SSIM_frame;
    SSIM_stack{1,4*(m_iter-1)+5} = SSIM_ROI_frame;

    resid_stack{m_iter}      = resid;
    outData_stack{m_iter}    = outData;
    lambdavali_stack(m_iter) = lambdavali;
    
    if m_iter == 1
        Ir_init = Ir_mc; % Keep first reconstruction for comparisson
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Save intermediate results
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    if SAVE_TRACES
        
        niter = size(xklog,1);
        nfram = length(optsf.logFrames);
        xklog = permute(reshape(xklog,[niter ROW COL nfram]), [2 3 4 1]);
        xklog = lpi_correct_csm_rms(xklog, csm.rms);
        xklog = reshape(xklog, [ROW COL*nfram niter]);
        
        if m_iter == 1
            varstosave = {'COL', 'ROW', 'N', 'Ir_mc', 'xklog','Ir_zero', ...
                'resid', 'outData', 'lambda_t', 'lambda_s', 'lambdavali'};
            sec2video([opts.out.traces num2str(m_iter-1) '_recon.avi'], Ir_zero);
        else
            ts = regout.ts; %#ok<NASGU>
            T_ite = regout.T_ite; %#ok<NASGU>
            varstosave = {'ts', 'T_ite', 'COL', 'ROW', ...
                'N', 'Ir_mc', 'xklog', 'resid', 'outData', ...
                'lambda_t', 'lambda_s', 'lambdavali'};
        end
        
        sec2video([opts.out.traces num2str(m_iter) '_evol.avi'],  xklog);
        sec2video([opts.out.traces num2str(m_iter) '_recon.avi'], Ir_mc);
                
        datafilename = strcat(opts.out.traces, num2str(m_iter,'%d'), '.mat');
        save(datafilename, varstosave{:});
    end
    
    fprintf('lambda_t = %g, lambda_s = %g, m_iter = %d. SER = %g, SER_ROI = %g\n',...
            lambda, lambda_s, m_iter, SER, SER_ROI);
end

exectimes.total = toc(tic_total);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Save results
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if SAVE_RESULTS
    % Main directory to save the data folders in
    outputresfile = opts.out.result;            

    trace.exectimes        = exectimes;
    trace.SER_stack        = SER_stack;
    trace.SSIM_stack       = SSIM_stack;
    trace.resid_stack      = resid_stack;
    trace.outData_stack    = outData_stack;
    trace.lambdavali_stack = lambdavali_stack;

    varstosave = {'I_full', 'Ir_mc', 'Ir_init', 'Ir_zero', 'trace'};
    save(outputresfile, varstosave{:});
    
    % Create video with results:
    Images_ALL = [I_full Ir_zero Ir_init Ir_mc];
    % movie2avi(sec2movie(Images_ALL),[outputresfile(1:end-4) '.avi']); % DEPRECATED
    sec2video([outputresfile(1:end-4) '.avi'], Images_ALL);

    % And create gif file
    sec2gif(Images_ALL, [outputresfile(1:end-4) '.gif'], round(1000/N)/1000);

end

end


function z = spTV(z, sm)
    vec = @(x)x(:);
    z = reshape(z,sm);
    z = [vec(diff(z,1,1)); vec(diff(z,1,2))];
end

function z = spTVt(z, sm)
    M = prod(sm-[1 0 0]);
    cv = reshape(z(1:M)    , sm - [1 0 0]);
    ch = reshape(z(M+1:end), sm - [0 1 0]);
    vec = @(x)x(:);

    z = vec([-cv(1,:,:); -diff(cv,1,1); cv(end,:,:)] + ...
        [-ch(:,1,:), -diff(ch,1,2), ch(:,end,:)]);
end

function z = tTV(z,sm)
    z = reshape(z,sm);
    z   = col(diff(z(:,:,[end 1:end]),1,3));
end

function z = tTVt(z,sm)
    z = reshape(z,sm);
    z = -col(diff(z(:,:,[1:end 1]),1,3));
end