function MASK = gauss_subsampling(N_ph, N_time, R, C_lines, sigma)
%GAUSS_SUBSAMPLING Random subsampling following a Gaussian profile.
% MASK = gauss_subsampling(N_ph, N_time, R, C_lines, sigma) generates
% random sampling MASK of size N_ph x N_time with sampling density
% following a Gaussian profile.
%    N_ph    : size along phase encoding dimension
%    N_time  : number of temporal frames
%    C_lines : Twice this number of central lines will be always acquired
%    sigma   : standard deviation of the Gaussian pattern
%
% Based on code from Salman Asif (with some modifications):
%  http://users.ece.gatech.edu/sasif/dynamicMRI/index.html
%
%   Copyright (C) 2015 Javier Royuela del Val <jroyval@lpi.tel.uva.es>
%   Laboratorio de Procesado de Imagen, Universidad de Valladolid
%   www.lpi.tel.uva.es
%
%   When indicated, some parts are based on code from Salman Asif:
%       http://users.ece.gatech.edu/sasif/dynamicMRI/index.html
%
%   This program is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
%
%   This program is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%
%   You should have received a copy of the GNU General Public License
%   along with this program.  If not, see <http://www.gnu.org/licenses/>.

if nargin < 5,
    sigma   = 1/3;
end

if nargin < 4,
    C_lines = 4;
end

N_rnd_lines = N_ph - 2 * C_lines;
N_target    = round(N_ph/R) - 2 * C_lines;

% Set a sigma. This will fix the width of the sampling pattern considering
% the k-space ranging from -1 to 1.
%    Note: if max() is used, pattern width will vary.
% Get N_target*N_time random samples between -1 and 1
% sigma = 1/3;
N_valid = 0;
idx1 = [];
while N_valid < N_target*N_time
    idx1  = [idx1; sigma * randn(N_target*N_time, 1 )]; %#ok<AGROW>
    idx1  = idx1(abs(idx1)<1);
    N_valid = length(idx1);
end
idx1 = reshape(idx1(1:N_target*N_time), [N_target N_time]);

% Displace positions and check
idx1  = ceil( idx1 * (N_rnd_lines / 2) ) + ceil( N_ph / 2 );
idx1(idx1<1) = 1;
idx1(idx1>N_rnd_lines) = N_rnd_lines;

MASK = false(N_rnd_lines, N_time);
for n=1:N_time
    MASK(idx1(:,n),n) = true;
    % Distribute the remaining samples uniformly
    residual = N_target - sum(MASK(:,n));
    if residual > 0
        temp_zeros = find(MASK(:,n)==0);
        m = length(temp_zeros);
        for j=1:residual
            temp = ceil(rand(1) * m);
            MASK(temp_zeros(temp),n)=true;
            temp_zeros(temp)=[];
            m = m - 1;
        end
    elseif residual < 0
        error('More samples than required');
    end
end

MASK = [MASK(1:floor(N_rnd_lines/2),:); ...
        true(2*C_lines,N_time); ...
        MASK(floor(N_rnd_lines/2)+1:end,:)];
    





              